import re
from datetime import date, datetime, timedelta

from workflow import MATCH_ALL, MATCH_ALLCHARS

from mstodo.models.preferences import Preferences, DEFAULT_TASKFOLDER_MOST_RECENT
from mstodo.util import parsedatetime_calendar, wf_wrapper

# Up to 8 words (sheesh!) followed by a colon
LIST_TITLE_PATTERN = re.compile(r'^((?:[^\s:]+ *){0,8}):',
                                re.UNICODE | re.IGNORECASE)

# The word "in" followed optionally by "list"
INFIX_LIST_KEYWORD_PATTERN = re.compile(r'\bin\s+(list\s+)?',
                                        re.UNICODE | re.IGNORECASE)

# `every N units` optionally preceded by `repeat`
RECURRENCE_PATTERN = re.compile(
    r'(?:\brepeat(?:ing|s)?:? )?(?:\bevery *(\d*) ' +
    r'*((?:day|week|month|year|d|w|m|y|da|wk|mo|yr)s?\b)?' +
    r'|(daily|weekly|monthly|yearly|annually))', re.UNICODE | re.IGNORECASE)
RECURRENCE_BY_DATE_PATTERN = re.compile(
    r'(?:\brepeat:? )?\bevery *((?:\S+ *){0,2})', re.UNICODE | re.IGNORECASE)

REMINDER_PATTERN = re.compile(
    r'(\b(?:remind me|reminder|remind|r|alarm)\b:? *)(.*)',
    re.UNICODE | re.IGNORECASE)

# Anything following the `due` keyword
DUE_PATTERN = re.compile(r'(\bdue:?\b\s*)(.*)', re.UNICODE | re.IGNORECASE)

NOT_DUE_PATTERN = re.compile(r'not? due( date)?', re.UNICODE | re.IGNORECASE)

HASHTAG_PROMPT_PATTERN = re.compile(r'(?:^|[\s:])(#\S*)$', re.UNICODE)

# An asterisk at the end of the phrase
STAR_PATTERN = re.compile(r'\*$', re.UNICODE)

# Tabs or multiple consecutive spaces
WHITESPACE_CLEANUP_PATTERN = re.compile(r'\t|\s{2,}', re.UNICODE)

# Split words ignoring leading and trailing punctuation
WORD_SEPARATOR_PATTERN = re.compile(r'\W*\s+\W*', re.UNICODE)

# Anything following the '//' delimiter
SLASHES_PATTERN = re.compile(r'( //)(.*)$', re.DOTALL)

# Maps first letter to the API recurrence type
RECURRENCE_TYPES = {
    'd': 'day',
    'w': 'week',
    'm': 'month',
    'y': 'year',
    # for "annually"
    'a': 'year'
}


class TaskParser():
    """
    Utility function to handle task creation and split out relevant content into title, note, recurrence, 
    reminder and so on.
    """
    phrase = None
    title = None
    list_id = None
    list_title = None
    due_date = None
    recurrence_type = None
    recurrence_count = None
    reminder_date = None
    hashtag_prompt = None
    assignee_id = None
    starred = False
    completed = False
    note = None

    has_list_prompt = False
    has_due_date_prompt = False
    has_recurrence_prompt = False
    has_reminder_prompt = False
    has_hashtag_prompt = False

    _list_phrase = None
    _due_date_phrase = None
    _recurrence_phrase = None
    _reminder_phrase = None
    _starred_phrase = None
    _note_phrase = None

    def __init__(self, phrase):
        self.phrase = phrase.strip()
        self._parse()

    def _parse(self):
        cls = type(self)
        phrase = self.phrase
        cal = parsedatetime_calendar()
        wf = wf_wrapper()
        taskfolders = wf.stored_data('taskfolders')
        prefs = Preferences.current_prefs()
        ignore_due_date = False

        match = re.search(HASHTAG_PROMPT_PATTERN, phrase)
        if match:
            self.hashtag_prompt = match.group(1)
            self.has_hashtag_prompt = True

        match = re.search(SLASHES_PATTERN, phrase)
        if match:
            self._note_phrase = match.group(1) + match.group(2)
            self.note = re.sub(
                WHITESPACE_CLEANUP_PATTERN, ' ', match.group(2)).strip()
            phrase = phrase[:match.start()] + phrase[match.end():]

        match = re.search(STAR_PATTERN, phrase)
        if match:
            self.starred = True
            self._starred_phrase = match.group()
            phrase = phrase[:match.start()] + phrase[match.end():]

        match = re.search(NOT_DUE_PATTERN, phrase)
        if match:
            ignore_due_date = True
            phrase = phrase[:match.start()] + phrase[match.end():]

        match = re.search(LIST_TITLE_PATTERN, phrase)
        if taskfolders and match:
            if match.group(1):
                matching_taskfolders = wf.filter(
                    match.group(1),
                    taskfolders,
                    lambda l: l['title'],
                    # Ignore MATCH_ALLCHARS which is expensive and inaccurate
                    match_on=MATCH_ALL ^ MATCH_ALLCHARS
                )

                # Take the first match as the desired list
                if matching_taskfolders:
                    self.list_id = matching_taskfolders[0]['id']
                    self.list_title = matching_taskfolders[0]['title']
            # The list name was empty
            else:
                self.has_list_prompt = True

            if self.list_title or self.has_list_prompt:
                self._list_phrase = match.group()
                phrase = phrase[:match.start()] + phrase[match.end():]

        # Parse and remove the recurrence phrase first so that any dates do
        # not interfere with the due date
        match = re.search(RECURRENCE_PATTERN, phrase)
        if match:
            type_phrase = match.group(2) if match.group(2) else match.group(3)
            if type_phrase:
                # Look up the recurrence type based on the first letter of the
                # work or abbreviation used in the phrase
                self.recurrence_type = RECURRENCE_TYPES[type_phrase[0].lower()]
                self.recurrence_count = int(match.group(1) or 1)
            else:
                match = re.search(RECURRENCE_BY_DATE_PATTERN, phrase)
                if match:
                    recurrence_phrase = match.group()
                    dates = cal.nlp(match.group(1), version=2)

                    if dates:
                        # Only remove the first date following `every`
                        datetime_info = dates[0]
                        # Set due_date if a datetime was found and it is not time only
                        if datetime_info[1].hasDate:
                            self.due_date = datetime_info[0].date()
                            date_expression = datetime_info[4]

                            # FIXME: This logic could be improved to better
                            # differentiate between week and year expressions

                            # If the date expression is only one word and the next
                            # due date is less than one week from now, set a
                            # weekly recurrence, e.g. every Tuesday
                            if len(date_expression.split(' ')) == 1 \
                                and self.due_date < date.today() + timedelta(days=8):
                                self.recurrence_count = 1
                                self.recurrence_type = 'week'
                            # Otherwise expect a multi-word value like a date,
                            # e.g. every May 17
                            else:
                                self.recurrence_count = 1
                                self.recurrence_type = 'year'

                            self.has_recurrence_prompt = False

                            # Pull in any words between the `due` keyword and the
                            # actual date text
                            date_pattern = re.escape(date_expression)
                            date_pattern = r'.*?' + date_pattern

                            # Prepare to set the recurrence phrase below
                            match = re.search(date_pattern, recurrence_phrase)

            # This is just the "every" keyword with no date following
            if not self.recurrence_type:
                self.has_recurrence_prompt = True

            self._recurrence_phrase = match.group()
            phrase = phrase.replace(self._recurrence_phrase, '', 1)


        reminder_info = None
        match = re.search(REMINDER_PATTERN, phrase)
        if match:
            datetimes = cal.nlp(match.group(2), version=2)

            # If there is at least one date immediately following the reminder
            # phrase use it as the reminder date
            if datetimes and datetimes[0][2] == 0:
                # Only remove the first date following the keyword
                reminder_info = datetimes[0]

                self._reminder_phrase = match.group(1) + reminder_info[4]
                phrase = phrase.replace(self._reminder_phrase, '', 1)
            # Otherwise if there is just a reminder phrase, set the reminder
            # to the default time on the date due
            else:
                # There is no text following the reminder phrase, prompt for a reminder
                if not match.group(2):
                    self.has_reminder_prompt = True
                self._reminder_phrase = match.group(1)

                # Careful, this might just be the letter "r" so rather than
                # replacing it is better to strip out by index
                phrase = phrase[:match.start(1)] + phrase[match.end(1):]


        due_keyword = None
        potential_date_phrase = None
        if not ignore_due_date:
            match = re.search(DUE_PATTERN, phrase)
            # Search for the due date only following the `due` keyword
            if match:
                due_keyword = match.group(1)

                if match.group(2):
                    potential_date_phrase = match.group(2)
            # Otherwise find a due date anywhere in the phrase
            elif not prefs.explicit_keywords:
                potential_date_phrase = phrase

        if potential_date_phrase:
            dates = cal.nlp(potential_date_phrase, version=2)

            if dates:
                # Only remove the first date following `due`
                datetime_info = dates[0]
                # Set due_date if a datetime was found and it is not time only
                if datetime_info[1].hasDate:
                    self.due_date = datetime_info[0].date()
                elif datetime_info[1].hasTime and not self.due_date:
                    self.due_date = date.today()

                if self.due_date:
                    # Pull in any words between the `due` keyword and the
                    # actual date text
                    date_pattern = re.escape(datetime_info[4])

                    if due_keyword:
                        date_pattern = re.escape(due_keyword) + r'.*?' + date_pattern

                    due_date_phrase_match = re.search(date_pattern, phrase)

                    if due_date_phrase_match:
                        self._due_date_phrase = due_date_phrase_match.group()
                        phrase = phrase.replace(self._due_date_phrase, '', 1)

                    # If the due date specifies a time, set it as the reminder
                    if datetime_info[1].hasTime:
                        if datetime_info[1].hasDate:
                            self.reminder_date = datetime_info[0]
                        elif self.due_date:
                            self.reminder_date = datetime.combine(self.due_date, datetime_info[0].time())
                # Just a time component
                else:
                    due_keyword = None
            # No dates in the phrase
            else:
                due_keyword = None

        # The word due was not followed by a date
        if due_keyword and not self._due_date_phrase:
            self.has_due_date_prompt = True
            self._due_date_phrase = match.group(1)

            # Avoid accidentally replacing "due" inside words elsewhere in the
            # string
            phrase = phrase[:match.start(1)] + phrase[match.end(1):]

        if self.recurrence_type and not self.due_date:
            self.due_date = date.today()

        if self._reminder_phrase:
            # If a due date is set, a time-only reminder is relative to that
            # date; otherwise if there is no due date it is relative to today
            reference_date = self.due_date if self.due_date else date.today()

            if reminder_info:
                (dt, datetime_context, _, _, _) = reminder_info

                # Date and time; use as-is
                if datetime_context.hasTime and datetime_context.hasDate:
                    self.reminder_date = dt
                # Time only; set the reminder on the due day
                elif datetime_context.hasTime:
                    self.reminder_date = cls.reminder_date_combine(reference_date, dt)
                # Date only; set the default reminder time on that day
                elif datetime_context.hasDate:
                    self.reminder_date = cls.reminder_date_combine(dt)

            else:
                self.reminder_date = cls.reminder_date_combine(reference_date)

        # Look for a list title at the end of the remaining phrase, like
        # "in list Office"
        if not self.list_title:
            matches = re.finditer(INFIX_LIST_KEYWORD_PATTERN, phrase)
            for match in matches:
                subphrase = phrase[match.end():]

                # Just a couple characters are too likely to result in a false
                # positive, but allow it if the letters are capitalized
                if len(subphrase) > 2 or subphrase == subphrase.upper():
                    matching_taskfolders = wf.filter(
                        subphrase,
                        taskfolders,
                        lambda f: f['title'],
                        # Ignore MATCH_ALLCHARS which is expensive and inaccurate
                        match_on=MATCH_ALL ^ MATCH_ALLCHARS
                    )

                    # Take the first match as the desired list
                    if matching_taskfolders:
                        self.list_id = matching_taskfolders[0]['id']
                        self.list_title = matching_taskfolders[0]['title']
                        self._list_phrase = match.group() + subphrase
                        phrase = phrase[:match.start()]
                        break

        # No list parsed, assign to Tasks
        if not self.list_title:
            if prefs.default_taskfolder_id and taskfolders:
                if prefs.default_taskfolder_id == DEFAULT_TASKFOLDER_MOST_RECENT:
                    self.list_id = prefs.last_taskfolder_id
                else:
                    self.list_id = prefs.default_taskfolder_id
                default_taskfolder = next((f for f in taskfolders if f['id'] == self.list_id), None)
                if default_taskfolder:
                    self.list_title = default_taskfolder['title']

            if not self.list_title:
                if taskfolders:
                    inbox = taskfolders[0]
                    self.list_id = inbox['id']
                    self.list_title = inbox['title']
                else:
                    self.list_id = 0
                    self.list_title = 'Tasks'

        # Set an automatic reminder when there is a due date without a
        # specified reminder
        if self.due_date and not self.reminder_date and prefs.automatic_reminders:
            self.reminder_date = cls.reminder_date_combine(self.due_date)

        # Condense extra whitespace remaining in the task title after parsing
        self.title = re.sub(WHITESPACE_CLEANUP_PATTERN, ' ', phrase).strip()

    @classmethod
    def reminder_date_combine(cls, date_component, time_component=None):
        """
        Returns a datetime based on the date portion of the date_component and
        the time portion of the time_component with special handling for an
        unspecified time_component. Based on the user preferences, a None
        time_component will result in either the default reminder time or an
        adjustment based on the current time if the reminder date is today.
        """
        prefs = Preferences.current_prefs()

        if isinstance(date_component, datetime):
            date_component = date_component.date()

        # Set a dynamic reminder date if due today
        if date_component == date.today() and time_component is None and prefs.reminder_today_offset:
            adjusted_now = datetime.now()
            adjusted_now -= timedelta(seconds=adjusted_now.second, microseconds=adjusted_now.microsecond)
            time_component = adjusted_now + prefs.reminder_today_offset_timedelta

            # "Round" to nearest 5 minutes, e.g. from :01 to :05, :44 to :45,
            # :50 to :50
            time_component += timedelta(minutes=(5 - time_component.minute % 5) % 5)

        # Default an unspecified time component on any day other than today to
        # the default reminder time
        if time_component is None:
            time_component = prefs.reminder_time

        if isinstance(time_component, datetime):
            time_component = time_component.time()

        return datetime.combine(date_component, time_component)

    def phrase_with(self, title=None, list_title=None, due_date=None,
                    recurrence=None, reminder_date=None, hashtag=None, starred=None):
        components = []
        #@TODO research this logic, assuming it is correct pythonic usage based on examples from other packages
        basestring = str

        # Retain the current list
        if list_title is None:
            if self._list_phrase:
                components.append(self._list_phrase)
        # Specifies a list by name
        elif isinstance(list_title, basestring):
            list_title = list_title.replace(':', '')
            components.append(list_title + ':')
        # Triggers selection of a list
        elif list_title:
            components.append(':')
        # Remove the current value
        else:
            pass

        # Add the task text
        if title:
            components.append(title)
        elif self.title:
            components.append(self.title)

        # Remove the hashtag prompt
        if hashtag and self.has_hashtag_prompt and (title or self.title):
            components[-1] = components[-1].rsplit(self.hashtag_prompt, 1)[0]

        # Retain the current due date
        if due_date is None:
            if self._due_date_phrase:
                components.append(self._due_date_phrase)
        # Specifies a due date phrase
        elif isinstance(due_date, basestring):
            components.append(due_date)
        # Triggers selection of a due date
        elif due_date:
            components.append('due ')
        # Remove the current value
        else:
            pass

        # Retain the current recurrence
        if recurrence is None:
            if self._recurrence_phrase:
                components.append(self._recurrence_phrase)
        # Specifies a recurrence phrase
        elif isinstance(recurrence, basestring):
            components.append(recurrence)
        # Triggers selection of a recurrence
        elif recurrence:
            components.append('every ')
        # Remove the current value
        else:
            pass

        # Retain the current reminder
        if reminder_date is None:
            if self._reminder_phrase:
                components.append(self._reminder_phrase)
        # Specifies a reminder phrase
        elif isinstance(reminder_date, basestring):
            components.append(reminder_date)
        # Triggers selection of a reminder
        elif reminder_date:
            components.append('remind me ')
        # Remove the current value
        else:
            pass

        # Add the hashtag
        if hashtag:
            if '#' not in hashtag:
                hashtag = '#' + hashtag
            components.append(hashtag)
        # Removes the star
        else:
            pass

        # Retain the current star status
        if starred is None:
            if self._starred_phrase:
                components.append(self._starred_phrase)
        # Adds a star
        elif starred:
            components.append('*')
        # Removes the star
        else:
            pass

        # Adds a note
        if self.note:
            components.append(self._note_phrase)

        # Remove any empty string components, often a blank title
        components = [component for component in components if component]

        phrase = ' '.join(components)
        phrase = re.sub(WHITESPACE_CLEANUP_PATTERN, ' ', phrase)

        return phrase
